//
//  Copyright (C) 2009-2012  Nick Gasson
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#ifndef INC_ITRACKSEGMENT_HPP
#define INC_ITRACKSEGMENT_HPP

#include "Platform.hpp"
#include "Maths.hpp"
#include "IXMLSerialisable.hpp"
#include "IMesh.hpp"

#include <vector>
#include <set>

// Types used for specifying track segments
namespace track {
   // TODO: This only needs (x, y) position and should contain
   // separate layer field - but will do for now
   typedef Point<int> Position;
   typedef Vector<int> Direction;

   // Uniquely identifies the location of a train and its orientation
   // along a piece of track
   // Used for verifying whether bits of track can join together
   typedef pair<Position, Direction> Connection;

   // Angles for curved track
   typedef int Angle;

   struct TravelToken;
   typedef function<void (const TravelToken&, double)> TransformFunc;
   typedef function<float (const TravelToken&, float)> GradientFunc;

   float flat_gradient_func(const TravelToken& t, float d);

   // Sums up all the information required to travel along a piece
   // of track
   struct TravelToken {
      // Direction of travel at /entry/
      Direction direction;

      // Position of entry
      Position position;

      // A function that transforms the location of the train so it will
      // render in the correct place for this track segment The functions
      // assumes that it is initially placed at the origin
      TransformFunc transformer;

      // A function that returns the gradient at any point
      GradientFunc gradientf;

      // Number of possible exits from this track segment given the direction
      // we are travelling in
      int num_exits;

      // Wrappers for the above functions

      void transform(float delta) const
      {
         transformer(*this, delta);
      }

      float gradient(float delta) const
      {
         return gradientf(*this, delta);
      }
   };
}

// Orientations for straight track
namespace axis {
   const track::Direction X = make_vector(1, 0, 0);
   const track::Direction Y = make_vector(0, 0, 1);
}

struct ITrackSegment;
typedef shared_ptr<ITrackSegment> ITrackSegmentPtr;

typedef vector<PointI> PointList;

// A segment of track which fits over a number of tiles
// Each track segment has an origin and one or more exits
struct ITrackSegment : IXMLSerialisable {
   virtual ~ITrackSegment() {}

   // Render the track with the origin in the centre
   virtual void render() const = 0;

   // Merge this track segment's mesh into a scenery mesh buffer
   virtual void merge(IMeshBufferPtr buf) const = 0;

   // Set the absolute position of the track in the world
   virtual void set_origin(int x, int y, float height) = 0;

   // Get the length of this track segment
   virtual float segment_length(const track::TravelToken& a_token) const = 0;

   // Get a travel token for this track segment starting at a particular
   // position and moving in a particular direciton
   virtual track::TravelToken get_travel_token(track::Position pos,
                                               track::Direction dir) const = 0;

   // True if a train can travel in this direction along the track
   virtual bool is_valid_direction(const track::Direction& dir) const = 0;

   // Return the position of the next segment of track and the
   // orientation of the train.
   // Note that this may not actually be a valid track segment!
   // You should call IMap::is_valid_track before using it (this
   // will happen, e.g. if the train runs off the end of the line)
   // The token passed here must have been generated by this track
   // segment when the train entered it.
   virtual track::Connection next_position(const track::TravelToken& a_token)
      const = 0;

   // Add all the endpoints of the track segment to the given list
   // Note that an endpoint is not the same as what is returned
   // from `next_position' - e.g. a straight track that takes up
   // one tile has a single endpoint which is its origin
   virtual void get_endpoints(PointList& list) const = 0;

   // Similar to endpoints, the `covers' of a track are the tiles
   // which are not endpoints but are underneath the track
   virtual void get_covers(PointList& output) const = 0;

   // Covers are tile vertices covered by the track segment
   virtual void get_height_locked(PointList& output) const = 0;

   // Add an exit to this section of track possibly generating
   // a new track segment
   // If this is not possible a null pointer is returned
   // If it can be merged a new track segment is returned that
   // may be bigger than the origin segment
   // The track may already have an exit here in which case
   // a pointer to itself will be returned
   virtual ITrackSegmentPtr merge_exit(PointI where,
                                       track::Direction dir) = 0;

   // Some track segments may have several states - e.g. points
   // These functions change the track state
   virtual bool has_multiple_states() const = 0;
   virtual void prev_state() = 0;
   virtual void next_state() = 0;

   // Set a hint to display something about the track state on the next render
   // call - e.g display an arrow over points
   virtual void set_state_render_hint() = 0;
};

ITrackSegmentPtr make_straight_track(const track::Direction& a_direction);
ITrackSegmentPtr make_crossover_track();
ITrackSegmentPtr make_points(track::Direction a_direction, bool reflect);
ITrackSegmentPtr make_slope_track(track::Direction axis, VectorF slope,
   VectorF slope_before, VectorF slope_after);
ITrackSegmentPtr make_spline_track(VectorI delta,
                                   track::Direction entry_dir,
                                   track::Direction exit_dir);

#endif
